@@ -1,16 +1,28 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
-__all__ = ("to_dataframe",)
+# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
+
+from __future__ import annotations
+
+from functools import reduce
+from operator import iconcat
 
 import awkward as ak
+from awkward._dispatch import high_level_function
 from awkward._nplikes.numpy import Numpy
-from awkward._nplikes.numpylike import NumpyMetadata
+from awkward._nplikes.numpy_like import NumpyMetadata
+
+__all__ = ("to_dataframe",)
 
 numpy = Numpy.instance()
 np = NumpyMetadata.instance()
 
 
+def _default_levelname(index: int) -> str:
+    return "sub" * index + "entry"
+
+
+@high_level_function()
 def to_dataframe(
-    array, *, how="inner", levelname=lambda i: "sub" * i + "entry", anonymous="values"
+    array, *, how="inner", levelname=_default_levelname, anonymous="values"
 ):
     """
     Args:
@@ -122,34 +134,56 @@ def to_dataframe(
               2         3.0  NaN
               3         4.0  NaN
     """
-    with ak._errors.OperationErrorContext(
-        "ak.to_dataframe",
-        {
-            "array": array,
-            "how": how,
-            "levelname": levelname,
-            "anonymous": anonymous,
-        },
-    ):
-        return _impl(array, how, levelname, anonymous)
+    # Dispatch
+    yield (array,)
+
+    # Implementation
+    return _impl(array, how, levelname, anonymous)
 
 
 def _impl(array, how, levelname, anonymous):
     try:
         import pandas
     except ImportError as err:
-        raise ak._errors.wrap_error(
-            ImportError(
-                """install the 'pandas' package with:
+        raise ImportError(
+            """install the 'pandas' package with:
 
     pip install pandas --upgrade
 
 or
 
     conda install pandas"""
-            )
         ) from err
 
+    def _merge_multi_indexed_dfs(
+        df1: pandas.DataFrame, df2: pandas.DataFrame, how: str
+    ) -> pandas.DataFrame:
+        """Merge two DataFrames after aligning their MultiIndex levels."""
+        # Extra levels. At most one of these will be non-empty.
+        df1_extra = tuple(n for n in df1.index.names if n not in df2.index.names)
+        df2_extra = tuple(n for n in df2.index.names if n not in df1.index.names)
+
+        # Reset extra levels to columns
+        if df1_extra:
+            df1 = df1.reset_index(level=df1_extra)
+        if df2_extra:
+            df2 = df2.reset_index(level=df2_extra)
+
+        # Merge on aligned indices
+        merged = pandas.merge(df1, df2, left_index=True, right_index=True, how=how)
+
+        # Fill missing values in extra levels and convert to int
+        for n in df1_extra:
+            merged[n] = merged[n].fillna(0).astype(int)
+        for n in df2_extra:
+            merged[n] = merged[n].fillna(0).astype(int)
+
+        # Set index back with extra levels
+        if df1_extra or df2_extra:
+            merged = merged.set_index(list(df1_extra) + list(df2_extra), append=True)
+
+        return merged
+
     if how is not None:
         out = None
         for df in to_dataframe(
@@ -158,7 +192,7 @@ or
             if out is None:
                 out = df
             else:
-                out = pandas.merge(out, df, how=how, left_index=True, right_index=True)
+                out = _merge_multi_indexed_dfs(out, df, how)
         return out
 
     def recurse(layout, row_arrays, col_names):
@@ -170,11 +204,10 @@ or
 
         elif layout.purelist_depth > 1:
             offsets, flattened = layout._offsets_and_flattened(axis=1, depth=1)
-            offsets = numpy.asarray(offsets)
-            starts, stops = offsets[:-1], offsets[1:]
+            starts, stops = offsets.data[:-1], offsets.data[1:]
             counts = stops - starts
             if ak._util.win or ak._util.bits32:
-                counts = layout.backend.index_nplike.astype(counts, np.int32)
+                counts = layout.backend.nplike.astype(counts, np.int32)
             if len(row_arrays) == 0:
                 newrows = [
                     numpy.repeat(numpy.arange(len(counts), dtype=counts.dtype), counts)
@@ -192,7 +225,8 @@ or
             if layout.is_union:
                 return [(ak.operations.to_numpy(layout), row_arrays, col_names)]
             else:
-                return sum(
+                return reduce(
+                    iconcat,
                     (
                         recurse(layout._getitem_field(n), row_arrays, (*col_names, n))
                         for n in layout.fields
@@ -201,7 +235,8 @@ or
                 )
 
         elif isinstance(layout, ak.contents.RecordArray):
-            return sum(
+            return reduce(
+                iconcat,
                 (
                     recurse(layout._getitem_field(n), row_arrays, (*col_names, n))
                     for n in layout.fields
@@ -212,7 +247,7 @@ or
         else:
             return [(ak.operations.to_numpy(layout), row_arrays, col_names)]
 
-    layout = ak.operations.to_layout(array, allow_record=True, allow_other=False)
+    layout = ak.operations.to_layout(array, allow_record=True, primitive_policy="error")
     if isinstance(layout, ak.record.Record):
         layout2 = layout.array[layout.at : layout.at + 1]
     else:
@@ -228,6 +263,17 @@ or
         else:
             columns = pandas.MultiIndex.from_tuples([col_names])
 
+        # Before filling, ensure dtype is wide enough for "nan" fill value. Need
+        # at least three characters / bytes for 'nan' or b'nan', respectively.
+        if numpy.ma.is_masked(column):
+            if np.issubdtype(column.dtype, np.str_):
+                char_width = column.dtype.itemsize // 4
+                if char_width < 3:
+                    column = column.astype(np.dtype(("U", 3)))
+            elif np.issubdtype(column.dtype, np.bytes_):
+                byte_width = column.dtype.itemsize
+                if byte_width < 3:
+                    column = column.astype(np.dtype(("S", 3)))
         # Pandas can't handle masked strings
         if np.issubdtype(column.dtype, np.str_):
             column = numpy.ma.filled(column, "nan")
@@ -319,9 +365,7 @@ def _union_to_record(unionarray, anonymous):
 
         missingarray = ak.contents.IndexedOptionArray(
             ak.index.Index64(
-                unionarray.backend.index_nplike.full(
-                    unionarray.length, -1, dtype=np.int64
-                )
+                unionarray.backend.nplike.full(unionarray.length, -1, dtype=np.int64)
             ),
             ak.contents.EmptyArray(),
         )
