@@ -13,8 +13,11 @@ to_list = ak.operations.to_list
 @pytest.fixture(scope="function", autouse=True)
 def cleanup_cuda():
     yield
+    try:
+        cp.cuda.Device().synchronize()  # wait for all kernels
+    except cp.cuda.runtime.CUDARuntimeError as e:
+        print("GPU error during sync:", e)
     cp._default_memory_pool.free_all_blocks()
-    cp.cuda.Device().synchronize()
 
 
 primes = [x for x in range(2, 1000) if all(x % n != 0 for n in range(2, x))]
@@ -844,7 +847,6 @@ def test_0115_generic_reducer_operation_mask():
     del array
 
 
-@pytest.mark.skip(reason="awkward_reduce_argmin is not implemented")
 def test_0115_generic_reducer_operation_ByteMaskedArray():
     content = ak.operations.from_iter(
         [
